/****************************************************************************
*
*  CDDBMMCD.CMD - update OS/2 PM CD-Player via CDDB
*
*  Copyright (C) 2001 by Marcel Mller
*  Release 0.34
*
*  This program is free software; you can redistribute it and/or
*  modify it under the terms of the GNU General Public License
*  as published by the Free Software Foundation; either version 2
*  of the License, or (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
****************************************************************************/

opt.version = 0.34

CALL RxFuncAdd 'SysLoadFuncs', 'RexxUtil', 'SysLoadFuncs'
CALL SysLoadFuncs

CALL RxFuncAdd 'MMLoadFuncs', 'RXMMUTL', 'MMLoadFuncs'
CALL MMLoadFuncs

NUMERIC DIGITS 12

/* SECTION 1 - main ********************************************************/

/* help */
ARG cmdl dummy
cmdl = TRANSLATE(cmdl, '-', '/')
IF cmdl = '-H' | cmdl = '-?' | cmdl = '--help' THEN DO
   SAY "CDDBMMCD - update OS/2 PM CD-Player via CDDB, version "opt.version
   SAY "Copyright (C) 2001 by Marcel Mller"
   SAY
   SAY "CDDBMMCD commandline options:"
   SAY "-p<profile>      Use specified profile instead of cddbmmcd.ini in current"
   SAY "                 directory."
   SAY "-d<device>       CDAUDIO device. Multiple devices may be seperated with komma."
   SAY "-l<path>         Path to a local CDDB."
   SAY "-l               Use local CDDB at default path."
   SAY "-l-              Disable local CDDB."
   SAY "-t<type>         Type of local CDDB. W[indows] or S[tandard] or [all]."
   SAY "-f<threshold>    Enable fuzzy search with specified threshold if no exact match"
   SAY "                 is found. Use -f0 to disable fuzzy search."
   SAY "-s<server>       Use specified cddb server(s). The separation character is komma."
   SAY "-s               Use default CDDB server."
   SAY "-s-              Disable CDDB server queries."
   SAY "-P<proxy>        Proxy-server."
   SAY "-u<user>         User name for CDDB HELLO."
   SAY "-h<host>         Host name for CDDB HELLO."
   SAY "-c<client>       Client name and version."
   SAY "-o[+|-]<opcodes> c  Check current CD if undefined."
   SAY "                 C  Check current CD always."
   SAY "                 q  Execute CDDB query (if required)."
   SAY "                 Q  Execute CDDB query of previous outstanding checks."
   SAY "                 o  Retry unsuccsessful queries, which are older than one month."
   SAY "                 O  Retry all unsuccsessful queries."
   SAY "                 u  Create upload packet for unsuccessful queries, where the"
   SAY "                    information is entered with the OS/2 CD-Player meanwhile."
   SAY "                    OPTION CURRENTLY NOT IMPLEMENTED!"
   SAY "                 U  Force creation of upload packet for currently inserted CD."
   SAY "                    OPTION CURRENTLY NOT IMPLEMENTED!"
   SAY "                 s  Upload packets."
   SAY "                    OPTION CURRENTLY NOT IMPLEMENTED!"
   SAY "-z<path>         Path for upload files. Default = .\"
   SAY "-w               Write specified options to profile as new defaults."
   EXIT
   END

/* parse command line */
opt.writeini = 0
opt.usedb = ''
nodb = ''
cmdl = ARG(1)
DO WHILE cmdl \= ''
   IF LEFT(cmdl, 1) = '"' THEN
      PARSE VAR cmdl '"'opt'"'cmdl
    ELSE
      PARSE VAR cmdl opt cmdl
   cmdl = STRIP(cmdl, 'l')
   IF LEFT(opt,1) \= '-' & LEFT(opt,1) \= '/' THEN
      CALL Error 10, 'Illegal option 'opt'.'
   /* option */
   optc = SUBSTR(opt, 2, 1)
   SELECT
    WHEN optc = 'p' THEN
      opt.profile = SUBSTR(opt, 3)
    WHEN LEFT(opt, 10) = '--profile=' THEN
      opt.profile = SUBSTR(opt, 11)
    WHEN optc = 'd' THEN
      opt.cddev = TRANSLATE(SUBSTR(opt, 3),, ',')
    WHEN LEFT(opt, 9) = '--device=' THEN
      opt.profile = TRANSLATE(SUBSTR(opt, 10),, ',')
    WHEN optc = 'l' THEN DO
      IF SUBSTR(opt,3) = '-' THEN
         nodb = nodb'L'
       ELSE; DO
         opt.usedb = opt.usedb'L'
         IF LENGTH(opt) > 2 THEN
            opt.localdb = ToPath(SUBSTR(opt, 3))
         END
      END
    WHEN opt = '--local' THEN
      opt.usedb = opt.usedb'L'
    WHEN opt = '--nolocal' THEN
      nodb = nodb'L'
    WHEN LEFT(opt, 8) = '--local=' THEN DO
      opt.usedb = opt.usedb'L'
      opt.localdb = ToPath(SUBSTR(opt, 9))
      END
    WHEN optc = 't' THEN DO
      tmp = TRANSLATE(SUBSTR(opt, 3))
      IF \ABBREV('AUTO', tmp) & \ABBREV('STANDARD', tmp) & \ABBREV('WINDOWS', tmp) THEN
         CALL Error 5, 'Unknown database type 'tmp'.'
      opt.dbtype = LEFT(tmp, 1, 'A')
      END
    WHEN LEFT(opt, 7) = '--type=' THEN DO
      tmp = TRANSLATE(SUBSTR(opt, 8))
      IF \ABBREV('AUTO', tmp) & \ABBREV('STANDARD', tmp) & \ABBREV('WINDOWS', tmp) THEN
         CALL Error 5, 'Unknown database type 'tmp'.'
      opt.dbtype = LEFT(tmp, 1, 'A')
      END
    WHEN optc = 's' THEN DO
      IF SUBSTR(opt,3) = '-' THEN
         nodb = nodb'S'
       ELSE; DO
         opt.usedb = opt.usedb'S'
         IF LENGTH(opt) > 2 THEN
            opt.dbserver = TRANSLATE(SUBSTR(opt, 3),, ',')
         END
      END
    WHEN opt = '--server' THEN
      opt.usedb = opt.usedb'S'
    WHEN opt = '--noserver' THEN
      nodb = nodb'S'
    WHEN LEFT(opt, 9) = '--server=' THEN DO
      opt.usedb = opt.usedb'S'
      opt.dbserver = TRANSLATE(SUBSTR(opt, 10),, ',')
      END
    WHEN optc = 'P' THEN
      opt.proxy = SUBSTR(opt, 3)
    WHEN LEFT(opt, 8) = '--proxy=' THEN
      opt.proxy = SUBSTR(opt, 9)
    WHEN optc = 'f' THEN DO
      opt.fuzzy = SUBSTR(opt, 3)
      IF \DATATYPE(opt.fuzzy, 'W') THEN
         CALL Error 5, 'Fuzzy threshold 'opt.fuzzy' is not a whole number.'
      END
    WHEN LEFT(opt, 8) = '--fuzzy=' THEN DO
      opt.fuzzy = SUBSTR(opt, 9)
      IF \DATATYPE(opt.fuzzy, 'W') THEN
         CALL Error 5, 'Fuzzy threshold 'opt.fuzzy' is not a whole number.'
      END
    WHEN optc = 'u' THEN
      opt.user = SUBSTR(opt, 3)
    WHEN LEFT(opt, 7) = '--user=' THEN
      opt.user = SUBSTR(opt, 8)
    WHEN optc = 'h' THEN
      opt.host = SUBSTR(opt, 3)
    WHEN LEFT(opt, 7) = '--host=' THEN
      opt.host = SUBSTR(opt, 8)
    WHEN optc = 'c' THEN
      opt.client = SUBSTR(opt, 3)
    WHEN LEFT(opt, 9) = '--client=' THEN
      opt.client = SUBSTR(opt, 10)
    WHEN optc = 'o' THEN DO
      opt.opcode = SUBSTR(opt, 3)
      tmp = VERIFY(opt.opcode, '+-cCqQoOuUs')
      IF tmp \= 0 THEN
         CALL Error 5, opt' contains illegal opcode 'SUBSTR(opt.opcode, tmp, 1)'.'
      END
    WHEN LEFT(opt, 5) = '--op=' THEN DO
      opt.opcode = SUBSTR(opt, 6)
      tmp = VERIFY(opt.opcode, '+-cCqQoOuUs')
      IF tmp \= 0 THEN
         CALL Error 5, opt' contains illegal opcode 'SUBSTR(opt.opcode, tmp, 1)'.'
      END
    WHEN optc = 'z' THEN
      opt.uplpath = SUBSTR(opt, 3)
    WHEN LEFT(opt, 13) = '--uploadpath=' THEN
      opt.uplpath = SUBSTR(opt, 14)
    WHEN optc = 'w' THEN
      opt.writeini = 1
    WHEN opt = '--write' THEN
      opt.writeini = 1
    OTHERWISE
      CALL Error 10, 'Illegal option 'opt'.'
      END
   END

IF opt.usedb = '' THEN
   DROP opt.usedb

/* split opcode command */
p = VERIFY(opt.opcode, '+-', 'M')
IF p \= 0 THEN DO
   opmod = SUBSTR(opt.opcode, p)
   opt.opcode = LEFT(opt.opcode, p-1)
   IF opt.opcode = '' THEN
      DROP opt.opcode
   END
/* read and update settings */
CALL ApplyDefault 'profile', ,                ,                DIRECTORY()||"\CDDBMMCD.INI"
CALL ApplyDefault 'cddev',   'CDAUDIO',       'CDAudioDevice'
CALL ApplyDefault 'localdb', '.\',            'LocalCDDBPath'
CALL ApplyDefault 'dbtype',  'ALL',           'LocalCDDBType'
CALL ApplyDefault 'fuzzy',    41,             'FuzzyThreshold'
CALL ApplyDefault 'dbserver','freedb.freedb.org','CDDBservers'
CALL ApplyDefault 'proxy',   '',              'ProxyServer'
CALL ApplyDefault 'user',    ,                'Username',      VALUE('USER',,'OS2ENVIRONMENT'), 'os2user'
CALL ApplyDefault 'host',    ,                'Hostname',      VALUE('HOSTNAME',,'OS2ENVIRONMENT')
CALL ApplyDefault 'client',  ,                'ClientString',  'cddbmmcd/2 'opt.version
CALL ApplyDefault 'opcode',  'cqQou',         'OperationCodes'
CALL ApplyDefault 'usedb',   'L',             'CDDBType'
CALL ApplyDefault 'cdpini',  ,                'CDPiniLocation',STRIP(VALUE('MMBASE',,'OS2ENVIRONMENT'),'T',';')'\CDP.INI'
CALL ApplyDefault 'uplpath', '.\',            'UploadDataPath'
/* postprocess opcode */
DO WHILE opmod \= ''
   p = VERIFY(opmod'+', '+-', 'M', 2)
   cmd = LEFT(opmod, 1)
   tmp = SUBSTR(opmod, 2, p -2)
   opmod = SUBSTR(opmod, p)
   IF cmd = '+' THEN
      DO i = 1 TO p-2
         IF POS(SUBSTR(tmp, i, 1), opt.opcode) = 0 THEN
            opt.opcode = opt.opcode||SUBSTR(tmp, i, 1)
         END
    ELSE
      DO i = p-2 TO 1
         opt.opcode = RemoveC(opt.opcode, SUBSTR(tmp, i, 1))
         END
   END
CALL ApplyDefault 'opcode',  , 'OperationCodes'
DO i = 1 TO LENGTH(nodb)
   opt.usedb = RemoveC(opt.usedb, SUBSTR(nodb, i, 1))
   END
CALL ApplyDefault 'usedb',   , 'CDDBType'

/* start ! */
CALL HandleCD
/* downloads */
CALL CDDBQuery

EXIT

/* check 4 uploads */
CALL CheckUpload
/* uploads */
CALL CDDBUpload

EXIT


/* SECTION 2 - high level functions ****************************************/

GetTOC: PROCEDURE
/* get TOC of CD

   parameters:
      ARG(1) MCI device

   return:
      'toc'  Table of contents of current CD
      ''     Error
*/
   TOC = ''
   /* open CDAUDIO */
   CALL MCIcmd 'open 'ARG(1)' alias c wait'
   /* get TOC */
   n = MCIcmd('status c number of tracks wait', 0)
   IF n >= 1 THEN DO
      CALL MCIcmd 'set c time format mmtime wait'
      TOC = n' 'MCIcmd('status c length wait') / 40
      DO i = 1 TO n
         TOC = TOC' 'MCIcmd('status c position track 'i' wait') / 40
         /*SAY i'S' MCIcmd('status c position track 'i' wait') / 40' 'i'L' MCIcmd('status c length track 'i' wait') / 40*/
         END
      TOC = TOC' 'WORD(TOC, n+2) + MCIcmd('status c length track 'n' wait') / 40
      END
   /* release device */
   CALL MCIcmd 'close c'
   RETURN TOC

HandleCD: PROCEDURE EXPOSE opt. cd.
/* handle currently inserted CD

   return:
      cd.0   number of cd entries
      cd.i   CD TOC
*/
   cd.0 = 0
   IF VERIFY('cCU', opt.opcode, 'M') = 0 THEN
      RETURN
   /* Multimedia-REXX-Untersttzung laden und initialisieren */
   CALL RXFUNCADD 'mciRxInit','MCIAPI','mciRxInit'
   CALL mciRxInit
   /* handle currently inserted CD(s) */
   devlist = opt.cddev
   SAY "Get table of contents of current CD"
   i = 1
   DO WHILE devlist \= ''
      PARSE VAR devlist dev','devlist
      cd.i = GetTOC(dev)
      IF cd.i \= '' THEN
         i = i +1
      END
   /* close MCI */
   CALL mciRxExit
   cd.0 = i -1
   RETURN

CDDBQuery: PROCEDURE EXPOSE opt. cd.
/* create CDDB query list and execute it

   parameters:
      cd.i   current cd's
*/
   IF VERIFY(opt.opcode, 'qQoO', 'M') = 0 THEN RETURN
   /* create query list */
   query.0 = 0
   /* current cd's */
   IF POS('q', opt.opcode) \= 0 THEN
      DO i = 1 TO cd.0
         state = GetCDState(cd.i)
         IF POS('C', opt.opcode) = 0 THEN DO
            IF ReadMMCD(cd.i) \= '' THEN DO
               /* data available */
               SAY "- well known"
               IF state = 1 THEN
                  CALL SetCDState cd.i, 0
               ITERATE
               END
            END
         /* query request */
         SAY "- CDDB query required"
         IF state = 1 THEN
            ITERATE
         CALL SetCDState cd.i, 1
         CALL AddCD cd.i
         END

   opq = POS('Q', opt.opcode) \= 0
   opo = POS('o', opt.opcode) \= 0
   IF POS('O', opt.opcode) \= 0 THEN
      opo = 2
   /* pending CDs */
   IF opq | (opo \= 0) THEN DO
      IF SysIni(opt.profile, 'PendingCDs', 'ALL:', 'list') \= '' THEN
         list.0 = 0
      time = UnixTime()
      DO i = 1 TO list.0
         tmp = GetCDState(list.i)
         SELECT
          WHEN tmp = 1 THEN
            IF \opq THEN ITERATE
          WHEN opo = 0 THEN
            ITERATE
          WHEN LEFT(tmp, 1) = 2 THEN
            IF opo = 1 & time < SUBSTR(tmp, 3) + 30*24*60*60 THEN ITERATE
          OTHERWISE ITERATE
            END
         CALL AddCD list.i
         END
      END
   /* execute query list */
   SAY "Execute "query.0" queries"
   DO i = 1 TO query.0
      CALL SetCDState query.i, ExecQuery(query.i)
      END
   RETURN

ExecQuery: PROCEDURE EXPOSE opt.
/* execute CDDB query

   parameters:
      ARG(1) TOC of CD to query

   return:
      2 time no success at time (unixtime)
      0      success
*/
   IF POS('L', opt.usedb) \= 0 THEN
      IF LocalQuery(ARG(1)) THEN RETURN 0
   IF POS('S', opt.usedb) \= 0 THEN DO
      CALL TCPIPInit
      DO j = 1 TO WORDS(opt.dbserver)
         IF ServerQuery(ARG(1), WORD(opt.dbserver, j)) THEN RETURN 0
         END
      END
   RETURN "2 "UnixTime()

LocalQuery: PROCEDURE EXPOSE opt.
/* request to local CDDB database

   parameters:
      ARG(1) table of contents of CD to query

   return:
      1      OK, found and stored
      0      not found or error
*/
   TOCn = WORD(ARG(1), 1)
   /* calc CD ID */
   s = 0
   DO i = 1 TO TOCn
      s = s + Qsum(TRUNC(WORD(ARG(1), i+2) / 75))
      END
   /* enhanced CD FIX !!!
      Length of last track of enhanced CDs is too long by aproximately 11000 frames */
   IF WORD(ARG(1), 2) > WORD(ARG(1), TOCn+3) THEN
      l = TRUNC((WORD(ARG(1), TOCn+3)-11000) / 75)-TRUNC(WORD(ARG(1), 3) / 75)
    ELSE
      l = TRUNC(WORD(ARG(1), TOCn+3) / 75)-TRUNC(WORD(ARG(1), 3) / 75)
   SAY l
   cdid = D2X(s//255, 2)D2X(l, 4)D2X(TOCn, 2)
   /* search for cddb record */
   SAY 'Local query for CD-ID: 'cdid
   CALL SysFileTree opt.localdb'*', 'dirs', 'DO'
   fuzzy = opt.fuzzy
   opt.fuzzy = 0
   DO i = 1 TO dirs.0
      IF ReadCDDBFile(dirs.i'\', cdid, ARG(1)) THEN RETURN 1 /* got it */
      END
   opt.fuzzy = fuzzy
   SAY '- no record for id 'cdid' found.'
   IF fuzzy = 0 | TOCn = 1 THEN RETURN 0
   /* try fuzzy search */
   lendelta = TRUNC((opt.fuzzy-1) / 75) + 2
   qsdelta = 10
   cdids = ""
   DO qs = s - qsdelta TO s + qsdelta
      DO len = l - lendelta TO l + lendelta
         cdids = cdids||D2X(qs//255, 2)D2X(len, 4)D2X(TOCn, 2)" "
         END
      END
   SAY "- trying fuzzy search ("opt.fuzzy")"
   DO UNTIL cdids = ""
      PARSE VAR cdids cdid cdids
      DO i = 1 TO dirs.0
         IF ReadCDDBFile(dirs.i'\', cdid, ARG(1)) THEN RETURN 1 /* got it */
         END
      END
   SAY '- no success.'
   RETURN 0

ReadCDDBFile: PROCEDURE EXPOSE opt.
/* read CDDB file and check if it matches TOC

   parameters:
      ARG(1) path including trailing '\'
      ARG(2) cdid (filename)
      ARG(3) table of contents

   return:
      1      OK, information stored
      0      error
*/
   file = ARG(1)ARG(2)
   IF opt.dbtype \= 'W' THEN
      /* try normal cddb */
      IF STREAM(file, 'c', 'open read') = 'READY:' THEN SIGNAL gotfile
   IF opt.dbtype = 'N' THEN RETURN 0
   /* try win version of cddb */
   c1 = LEFT(ARG(2),1)
   c2 = SUBSTR(ARG(2), 2, 1)
   file = ARG(1)||c1||TRANSLATE(c2,"02468ACE","13579BDF")"to"c1||TRANSLATE(c2,"13579BDF","02468ACE")
   IF STREAM(file, 'c', 'open read') \= 'READY:' THEN RETURN 0
gotfile:
   r = ParseCDDBFile('l = LINEIN("'file'")', ARG(2), ARG(3));
   CALL STREAM file, 'c', 'close'
   RETURN r

ParseCDDBFile: PROCEDURE EXPOSE lines opt.
/* parse cddb file and store infos

   parameters:
      ARG(1) readline command, should store result in l
      ARG(2) cdid (filename)
      ARG(3) table of contents

   return:
      1      OK, information stored
      0      error
*/
   TOCn = WORD(ARG(3), 1)
   DO i = 0 TO TOCn
      TIT.i = "" /* DROP ist NOT sufficient (perl would be nice) */
      END
   state = 1
   DO line = 0
      /* read line */
      INTERPRET ARG(1)
      IF l = '' THEN LEAVE
      SELECT
       WHEN LEFT(l, 10) = "#FILENAME=" THEN DO
         IF state = 2 THEN LEAVE line /* faster finish */
         state = TRANSLATE(SUBSTR(l, 11)) = ARG(2)
         line = 0
         END
       WHEN state = 0 THEN NOP
       WHEN state = 1 THEN DO
         IF LEFT(l, 6) \= '# xmcd' THEN DO
            IF line < 5 THEN ITERATE line
            SAY 'cddb file 'file' has an unknown format.'
            RETURN 0
            END
         state = 2
         END
       WHEN LEFT(l, 22) = "# Track frame offsets:" THEN DO
         /* read and check track offsets */
         IF opt.fuzzy \= 0 THEN DO
            INTERPRET ARG(1)
            PARSE VAR l '#' offset
            diff = STRIP(TRANSLATE(offset,,D2C(9))) - WORD(ARG(3), 3)
            DO i = 2 TO TOCn
               INTERPRET ARG(1)
               PARSE VAR l '#' offset
               offset = STRIP(TRANSLATE(offset,,D2C(9)))
               IF ABS(WORD(ARG(3), i+2) + diff - offset) >= opt.fuzzy THEN DO
                  SAY ARG(2) "did not match: Track "i", "offset", "WORD(ARG(3), i+2)", "diff
                  RETURN 0
                  END
               END
            END
          ELSE DO i = 1 TO TOCn
            INTERPRET ARG(1)
            PARSE VAR l '#' offset
            offset = STRIP(TRANSLATE(offset,,D2C(9)))
            IF offset \= WORD(ARG(3), i+2) THEN DO
               SAY ARG(2) "did not match: Track "i", "offset", "WORD(ARG(3), i+2)
               RETURN 0
               END
            END
         /* well, the number of tracks should match */
         END
       WHEN LEFT(l, 1) = '#' THEN NOP /* any other comment */
       OTHERWISE
         PARSE VAR l item'='val
         SELECT
          WHEN item = 'DTITLE' THEN TIT.0 = TIT.0" "STRIP(val)
          WHEN LEFT(item,6) = 'TTITLE' THEN DO
            i = SUBSTR(item,7) + 1
            TIT.i = TIT.i" "STRIP(val)
            END
          OTHERWISE /* this should be superflous - well, not in my case W4,FP15 */
            END
         END
      END
   IF state = 0 THEN RETURN 0 /* win cddb: no entry found */
   title = SUBSTR(TIT.0, 2)
   DO i = 1 TO TOCn
      title = title'00'x||SUBSTR(TIT.i, 2)
      END
   CALL StoreMMCD ARG(3), title
   RETURN 1

ServerQuery: PROCEDURE EXPOSE opt.
/* request to CDDB server

   parameters:
      ARG(1) table of contents of CD to query
      ARG(2) cddb server

   return:
      1      OK, found and stored
      0      not found or error
*/
   /* bounce any server queries if one of the environment vars is missing */
   IF opt.user = '' | opt.host = '' THEN DO
      SAY "One of the following parameters is missing:"
      SAY " host name (%HOSTNAME% or -h) or"
      SAY " user name (%USER% or -u)."
      RETURN 0
      END

   r = 0
   tos = WORDINDEX(ARG(1),3)
   toe = LASTPOS(' ',ARG(1))
   cddbtoc = WORD(ARG(1), 1)' 'SUBSTR(ARG(1), tos, toe-tos)' 'TRUNC(SUBSTR(ARG(1), toe+1)/75)

   IF CDDBinit(ARG(2)) = '' THEN RETURN 0

   tmp = CDDBCommand('DISCID 'cddbtoc, 2)
   IF tmp = '' THEN RETURN CDDBExit(0)
   discid = WORD(tmp, WORDS(tmp))

   tmp = CDDBCommand('CDDB QUERY 'discid' 'cddbtoc, 2)
   SAY '- 'tmp
   SELECT
    WHEN WORD(tmp,1) = 200 THEN DO
      PARSE VAR tmp reply cat discid dummy
      lines = CDDBCommand('CDDB READ 'cat' 'discid, 2)
      IF WORD(lines,1) = 210 THEN
         r = ParseCDDBFile("PARSE VAR lines l'0D0A'x lines", discid, ARG(1))
      END
    WHEN WORD(tmp,1) = 211 THEN DO
      PARSE VAR tmp dummy'0D0A'x matches
      IF WORD(ARG(1), 1) <= 2 THEN
         SAY '- inexcat match with <= 2 tracks: giving up'
       ELSE
         DO WHILE matches \= '' & r = 0
            PARSE VAR matches tmp'0D0A'x matches
            PARSE VAR tmp cat discid dummy
            IF dummy = '' THEN LEAVE
            lines = CDDBCommand('CDDB READ 'cat' 'discid, 2)
            IF WORD(lines,1) = 210 THEN
               r = ParseCDDBFile("PARSE VAR lines l'0D0A'x lines", discid, ARG(1))
            END
      END
    OTHERWISE
      SAY '- not found'
      END

   RETURN CDDBExit(r)

/* CDDB protocol ... */
CDDBInitcddb: PROCEDURE EXPOSE so opt.
/* Initialize CDDB server access

   parameter:
      ARG(1)    server

   return:
      ''        error
      otherwise OK
*/
   /* parse server */
   PARSE VALUE ARG(1) WITH server':'port
   IF port = '' THEN port = 8880

   addr.family = 'AF_INET'
   IF VERIFY(server, '0123456789.') = 0 THEN
      addr.addr = server
    ELSE /* DNS lookup */
      IF SockGetHostByName(server, 'addr.') \= 1 THEN DO
         SAY 'DNS lookup for 'server' failed.'
         RETURN 0
         END
   addr.port = port
   so = SockSocket('AF_INET', 'SOCK_STREAM', 'IPPROTO_TCP')
   IF so < 0 THEN
      CALL Error 10, 'Failed to create socket'
   IF SockConnect(so, 'addr.') \= 0 THEN DO
      SAY 'Failed to connect 'server':'port
      RETURN ''
      END
   IF SockSetSockOpt(so, 'SOL_SOCKET', 'SO_RCVTIMEO', 30) \= 0 THEN DO
      SAY 'Failed to set socket options ('server':'port')'
      RETURN ''
      END
   SAY 'cddb://'server':'port
   SAY '- 'CDDBCommandcddb(,2)

   CALL CDDBCommandcddb 'CDDB HELLO 'opt.user' 'opt.host' 'opt.client, 2
   RETURN 1

CDDBCommandcddb: PROCEDURE EXPOSE so
/* Send CDDB command and wait for reply

   parameter:
      ARG(1) command (optional)
      ARG(2) maximum reply message class (optional), <0 means read until '.'

   return:
      ''     error
      cddb server reply
*/
   IF SYMBOL('so') \= 'VAR' THEN RETURN ''
   IF ARG(1,'e') THEN DO
      IF SockSend(so, ARG(1)'0A'x) \= LENGTH(ARG(1))+1 THEN DO
         SAY 'Failed to send "'ARG(1)'" to server.'
         CALL SockSoClose(so)
         DROP so
         RETURN ''
         END
      END
   IF SockRecv(so, 'tmp', 32000) = -1 THEN DO
      SAY 'Failed to receive data'
      CALL SockSoClose(so)
      DROP so
      RETURN ''
      END
   IF SUBSTR(tmp, 2) = 1 THEN
      DO WHILE Pos('0D0A2E0D0A'x, RIGHT(tmp, 10)) = 0
         IF SockRecv(so, 'tmp2', 32000) = -1 THEN DO
            SAY 'Failed to receive data 2'
            CALL SockSoClose(so)
            DROP so
            RETURN ''
            END
         tmp = tmp||tmp2
         END
   tmp = STRIP(STRIP(tmp,,'0A'x),,'0D'x)
   IF ARG(2, 'e') & LEFT(tmp, 1) > ARG(2) THEN DO
      SAY 'Unexpected CDDB server reply:'
      SAY ' > 'ARG(1)
      SAY ' < 'tmp
      CALL SockSoClose(so)
      DROP so
      RETURN ''
      END
   RETURN tmp

CDDBExitcddb: PROCEDURE EXPOSE so
   IF SYMBOL('so') = 'VAR' THEN
      CALL SockSoClose(so)
   RETURN ARG(1)

/* http protocol ... */
CDDBInithttp: PROCEDURE EXPOSE cgiScript addr. opt.
/* Initialize CDDB server access

   parameter:
      ARG(1)    server

   return:
      ''        error
      otherwise OK
*/
   /* use proxy server if set */
   IF opt.proxy \= '' THEN DO
      cgiScript = 'http://'ARG(1)'/~cddb/cddb.cgi'
      PARSE VAR opt.proxy server':'prt
      END
    ELSE DO
      cgiScript = '/~cddb/cddb.cgi'
      PARSE VALUE ARG(1) WITH server':'prt
      END
   IF prt = '' THEN prt = 80

   addr.family = 'AF_INET'
   IF VERIFY(server, '0123456789.') = 0 THEN
      addr.addr = server
    ELSE /* DNS lookup */
      IF SockGetHostByName(server, 'addr.') \= 1 THEN DO
         SAY 'DNS lookup of 'server' failed.'
         RETURN ''
         END
   addr.port = prt
   SAY 'POST 'cgiScript' -> 'addr.addr':'prt

   RETURN 1

CDDBCommandhttp: PROCEDURE EXPOSE opt. cgiScript addr.
/* Send HTTP CDDB command and wait for reply

   parameter:
      ARG(1) cddb command
      ARG(2) maximum reply message class (optional)

   return:
      ''     error
      cddb server reply
*/
   so = SockSocket('AF_INET', 'SOCK_STREAM', 'IPPROTO_TCP')
   IF so < 0 THEN
      CALL Error 10, 'Failed to create socket'
   IF SockConnect(so, 'addr.') \= 0 THEN DO
      SAY 'Failed to connect 'addr.addr':'addr.port
      CALL SockSoClose(so)
      RETURN ''
      END
   /* set a reasonable timeout for HTTP transactions */
   IF SockSetSockOpt(so, 'SOL_SOCKET', 'SO_RCVTIMEO', 30) \= 0 THEN DO
      SAY 'Failed to set socket options ('addr.addr':'addr.port')'
      CALL SockSoClose(so)
      RETURN ''
      END

   cddbHello = opt.user' 'opt.host' 'opt.client
   cddbMsg = 'cmd='TRANSLATE(ARG(1),'+',' ')'&hello='TRANSLATE(cddbHello,'+',' ')'&proto=5'
   postHdr = 'User-Agent: 'opt.client||'0D0A'x||'Accept: */*'||'0D0A'x||'Content-length: 'LENGTH(cddbMsg)'0D0A'x
   httpCmd = 'POST 'cgiScript' HTTP/1.0'||'0D0A'x||postHdr'0D0A'x||cddbMsg||'0D0A0D0A'x
   IF SockSend(so, httpCmd) \= LENGTH(httpCmd) THEN DO
      SAY 'Failed to send "'httpCmd'" to server.'
      CALL SockSoClose(so)
      RETURN ''
      END
   IF SockRecv(so, 'tmp', 32000) = -1 THEN DO
      SAY 'Failed to receive data'
      CALL SockSoClose(so)
      RETURN ''
      END
   DO WHILE Pos('0D0A'x, RIGHT(tmp, 2)) = 0
      IF SockRecv(so, 'tmp2', 32000) = -1 THEN DO
         SAY 'Failed to receive data 2'
         CALL SockSoClose(so)
         RETURN ''
         END
      tmp = tmp||tmp2
      END
   tmp = SUBSTR(tmp, POS('0D0A0D0A'x,tmp)+4)
   IF ARG(2,'e') & LEFT(tmp,1) > ARG(2) THEN DO
      SAY 'Unexpected CDDB server reply:'
      SAY ' > 'httpCmd
      SAY ' < 'tmp
      CALL SockSoClose(so)
      RETURN ''
      END
   CALL SockSoClose(so)
   RETURN tmp

CDDBExithttp:
   RETURN ARG(1)

/* dispatchers ... */
CDDBInit:
/* Prepare CDDB query

   parameter:
      ARG(1) server URL

   return:
      ''        error
      otherwise OK
*/
   /* parse protocol */
   PARSE VALUE TRANSLATE(ARG(1),'/','\') WITH protocol '://' server
   IF server = '' THEN DO
      protocol = 'cddb'
      server = ARG(1)
      END
    ELSE IF protocol = 'cddbp' THEN
      protocol = 'cddb'
    ELSE IF protocol \= 'cddb' & protocol \= 'http' THEN DO
      SAY 'Unknown cddb protocol: 'protocol
      RETURN ''
      END
   /* protocol dispatcher */
   INTERPRET 'RETURN CDDBInit'protocol'(server)'

CDDBCommand: /* Protocol dispatcher */
   IF ARG(2, 'e') THEN
      INTERPRET 'RETURN CDDBCommand'protocol'(ARG(1),ARG(2))'
    ELSE
      INTERPRET 'RETURN CDDBCommand'protocol'(ARG(1))'

CDDBExit: /* Protocol dispatcher */
   INTERPRET 'RETURN CDDBExit'protocol'(ARG(1))'


CheckUpload: PROCEDURE EXPOSE opt. cd.
/* check for uploads

   parameters:
      cd.i   table of contents of current CDs
*/
   /* create update list */
   query.0 = 0
   /* currect CDs */
   IF POS('U', opt.opcode) \= 0 THEN
      DO i = 1 TO cd.0
         CALL AddCD cd.i, ReadMMCD(cd.i)
         END
   /* old unsuccessful queries */
   IF POS('u', opt.opcode) \= 0 THEN DO
      IF SysIni(opt.profile, 'PendingCDs', 'ALL:', 'list') \= '' THEN
         list.0 = 0
      DO i = 1 TO list.0
         tmp = GetCDState(list.i)
         SELECT
          WHEN tmp = 3 THEN
            info = ReadMMCD(list.i)
          WHEN LEFT(tmp, 1) = 2 THEN DO
            /* check for data */
            info = ReadMMCD(list.i)
            IF info = '' THEN ITERATE
            SAY "- createing upload"
            CALL SetCDState list.i, 3
            END
          OTHERWISE ITERATE
            END
         CALL AddCD list.i, info
         END
      END
   /* create upload packets */
   SAY "Create "query.0" upload packets"
   DO i = 1 TO query.0
      CALL SetCDState query.i, CreateUpload(query.i, data.i)
      END
   RETURN

CreateUpload: PROCEDURE EXPOSE opt.
/* create upload packet

   parameters:
      ARG(1) TOC of CD
      ARG(2) cd-title||'0'x||track1||'0'x||...trackn
      ARG(3) revision, optional

   return:
      4 file OK, packet created, stored to file
      3      error
*/
   /* create file */
   fname = opt.uplpath
   RETURN 3

CDDBUpload: PROCEDURE EXPOSE opt.
/* execute any upload packets
*/
   IF POS('s', opt.opcode) = 0 THEN RETURN
   /* create upload list */
   query.0 = 0
   IF SysIni(opt.profile, 'PendingCDs', 'ALL:', 'list') \= '' THEN
      list.0 = 0
   DO i = 1 TO list.0
      tmp = GetCDState(list.i)
      IF LEFT(tmp, 1) \= 4 THEN ITERATE
      CALL AddCD SUBSTR(tmp, 3)
      END
   /* upload! */
   SAY "Execute "query.0" uploads"
   DO i = 1 TO query.0
      CALL SetCDState query.i, ExecUpload(query.i)
      END

ExecUpload: PROCEDURE EXPOSE opt.
/* execute upload

   parameters:
      ARG(1) filename

   return:
      0      OK, done
      4 file error
*/
   RETURN 4 ARG(1)


CalcMMCD: PROCEDURE
/* calculate CD key of MMPM CD player

   parameter:
      ARG(1) table of contents (from GetTOC)

   return:
      mmcdkey
*/
   n = WORD(ARG(1),1)
   RETURN TRANSLATE('1245',F2MSF(WORD(ARG(1),2)),'12345')TRANSLATE('1245.78',F2MSF(WORD(ARG(1),2)-WORD(ARG(1),n+3)+WORD(ARG(1),n+2)),'12345678')

ReadMMCD: PROCEDURE EXPOSE opt.cdpini
/* check if CD information is in CDP.INI

   parameter:
      ARG(1) table of contents

   return:
      cd-title||'0'x||track1||'0'x||...trackn
      ''     no or incomplete information found
*/
   mmcdkey = CalcMMCD(ARG(1))
   SAY "ReadMMCD: "mmcdkey
   /* read data */
   res = SysIni(opt.cdpini, mmcdkey, 'IMMCDDiscTitle')
   IF res = 'ERROR:' THEN RETURN ''
   DO i = WORD(ARG(1), 1) TO 1 BY -1
      tmp = SysIni(opt.cdpini, mmcdkey, i)
      IF tmp = 'ERROR:' THEN RETURN ''
      res = res'00'x||tmp
      END
   RETURN res

StoreMMCD: PROCEDURE EXPOSE opt.cdpini
/* store CD information in CDP.INI

   parameters:
      ARG(1) table of contents
      ARG(2) cd-title||'0'x||track1||'0'x||...trackn
*/
   IF ARG(2) = '' THEN RETURN
   /* calc MMCD key */
   mmcdkey = CalcMMCD(ARG(1))
   /* store data */
   PARSE VALUE Cp2DOS(ARG(2)) WITH tmp'00'x titles
   CALL SysIni opt.cdpini, mmcdkey, 'IMMCDDiscTitle', tmp
   SAY '- information stored: 'tmp
   DO i = 1 UNTIL titles = ''
      PARSE VAR titles tmp'00'x titles
      CALL SysIni opt.cdpini, mmcdkey, i, tmp
      END
   RETURN

GetCDState: PROCEDURE EXPOSE opt.profile
/* query CD state from profile

   parameters:
      ARG(1) table of contents

   return:
      0      unknown
      1      query request
      2 time not found at 'time' (unixtime)
      3      upload request
      4      upload pending
*/
   state = SysIni(opt.profile, 'PendingCDs', ARG(1))
   IF state = 'ERROR:' THEN
      state = 0
   RETURN state

SetCDState: PROCEDURE EXPOSE opt.profile
/* update CD state from profile

   parameters:
      ARG(1) table of contents
      ARG(2) new state
*/
   /*SAY "SetState: "ARG(1) ARG(2)*/
   val = ARG(2)
   IF val = 0 THEN
      val = 'DELETE:'
   RETURN SysIni(opt.profile, 'PendingCDs', ARG(1), val) = ''

/* SECTION 3 - helper functions ********************************************/

TCPIPInit: PROCEDURE
/* initialize socket functions */
   CALL RxFuncAdd "SockLoadFuncs","rxSock","SockLoadFuncs"
   CALL SockLoadFuncs
   IF SockInit() \= 0 THEN
      CALL Error 99, "Failed to initialize Socket API."
   RETURN

MCIcmd: PROCEDURE /* execute MCI command and return result */
   /*SAY ">"ARG(1)">"*/
   rc = mciRxSendString(ARG(1), 'mcir', 0, 0)
   IF rc \= 0 & ARG(2) \= 0 THEN DO
      CALL mciRxGetErrorString rc, 'errstr'
      SAY "MCI Error "rc" - "errstr
      SAY "Command: "ARG(1)
      SAY "Result: "mcir
      CALL mciRxSendString 'close c', 'mcir', 0, 0
      CALL mciRxExit
      EXIT 255
      END
   /*SAY "<"mcir"<"*/
   RETURN mcir

UnixTime: PROCEDURE
/* return unix time
   However, the time zone is ignored here
*/
   RETURN (DATE('B')-719162)*24*60*60 + TIME('S')

Qsum: PROCEDURE /* Quersumme */
   r = 0
   DO i = 1 TO LENGTH(ARG(1))
      r = r + SUBSTR(ARG(1), i, 1)
      END
   RETURN r

F2MSF: PROCEDURE /* frames -> mm:ss:ff */
   s = TRUNC(ARG(1)/75)
   f = ARG(1) - s*75
   m = TRUNC(s/60)
   s = s - m*60
   RETURN RIGHT(m,2,'0')':'RIGHT(s,2,'0')':'RIGHT(f,2,'0')

RemoveC: PROCEDURE /* remove character from string */
   p = POS(ARG(2), ARG(1))
   IF p = 0 THEN
      RETURN ARG(1)
   RETURN DELSTR(ARG(1), p, LENGTH(ARG(2)))

ToPath: /* preprocess path setting */
   RETURN STRIP(TRANSLATE(ARG(1), '\', '/'), 't', '\')'\'

ApplyDefault: PROCEDURE EXPOSE opt.
/* apply default value if value not set yet
   prefer profile setting
   write back

   parameters:
      ARG(1) name of the option
      ARG(2) default value
      ARG(3) name of the profile entry (if any)
      ARG(4...) dynamic default values (if field is still empty)
*/
   IF SYMBOL('opt.'ARG(1)) = 'VAR' THEN
      INTERPRET 'tmp = opt.'ARG(1)
    ELSE DO
      IF ARG(3,'e') THEN DO
         /* get setting from profile */
         tmp = SysIni(opt.profile, 'Settings', ARG(3))
         IF tmp = 'ERROR:' THEN DROP tmp
         END
      IF SYMBOL('tmp') = 'LIT' THEN
         tmp = ARG(2)
      END
   IF opt.writeini & ARG(3,'e') THEN
      CALL SysIni opt.profile, 'Settings', ARG(3), tmp
   DO i = 4 WHILE tmp = '' & ARG(i,'e')
      tmp = ARG(i)
      END
   INTERPRET 'opt.'ARG(1)' = "'tmp'"'
   RETURN

AddCD: PROCEDURE EXPOSE opt. query. data.
/* add TOC entry to CD list

   parameters:
      ARG(1) table of contents from CD
      ARG(2) data to store
*/
   IF ARG(1) = '' | ARG(1) = query.1 THEN
      RETURN /* empty or duplicate */
   tmp = query.0 +1
   query.tmp = ARG(1)
   query.0 = tmp
   IF ARG(2, 'e') THEN
      data.tmp = ARG(2)
   RETURN

Cp2DOS: /* CP translation ISO to current CP */
   RETURN MMTranslateCp(ARG(1), 1004, , '_')

Error: /* raise error */
   SAY ARG(2)
   EXIT ARG(1)

